/**
 * Header file of TCB services C library.
 *
 * author: Paulo Sousa (pjsousa@di.fc.ul.pt)
 * date: 21/03/2003
 */


// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the TCB_SERVICES_LIBRARY_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// TCB_SERVICES_LIBRARY_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.
#ifdef TCB_SERVICES_LIBRARY_EXPORTS
#define TCB_SERVICES_LIBRARY_API __declspec(dllexport)
#else
#define TCB_SERVICES_LIBRARY_API __declspec(dllimport)
#endif


/**
 * Only used for tests
 */
extern TCB_SERVICES_LIBRARY_API int nTcb_services_library;
TCB_SERVICES_LIBRARY_API int fnTcb_services_library(void);


//we need this for NUMBER_OF_MACHINES
#include "../tcb_wince/config_parameters.h"

//number of bytes used to transport TCB info in each packet sent through a TCB socket.
#define TCB_DATA_SIZE 9


/**
 * Info returned by the local measurement service
 */
typedef struct tcb_measurement_info {
	//end instant
	int end_ev;
	//measurement value
	int procdelay;
} tcb_measurement_info_t;


/**
 * Info returned by the local failure detection service
 */
typedef struct tcb_local_fd_info {
	//end instant
	int end_ev;
	//time elapsed between failure detection start instant and end_ev
	int procdelay;
	//indicates if a timing failure occured
	bool faulty;
} tcb_local_fd_info_t;


/**
 * Info about the result of a distributed timing failure detection in a machine.
 */
typedef struct tcb_distributed_fail_info {
	//indicates if a failure occured:
	//0-no failure
	//1-failure
	//2-no info
	char fail;
	//indicates if lateness is available
	bool latenessAvail;
	//indicates the lateness degree if it's available
	int latenessDegree;
} tcb_distributed_fail_info_t;


/**
 * Info returned by wait_info. It contains information about the result of
 * a distributed timing failure detection.
 */
typedef struct tcb_distributed_fd_info {
	//information sender id
	char from;
	//information sequence number
	int seq;
	//info about failure in each machine
	tcb_distributed_fail_info_t failInfo[NUMBER_OF_MACHINES];
} tcb_distributed_fd_info_t;


/**
 * One of the two possible infos returned by tcb_recvfrom. 
 * It contains information about the result of a distributed measurement.
 */
typedef struct tcb_distributed_measurement_info {
	//instant of message reception
	int delivery_ev;
	//duration of the action terminated at delivery_ev
	int delay;
	//measurement error (ms)
	int error;
} tcb_distributed_measurement_info_t;


/**
 * One of the two possible infos returned by tcb_recvfrom. 
 * It contains the id of distributed failure detection.
 */
typedef struct tcb_distributed_fd_id {
	//id of the TCB associated with the process that requested the service
	char src_id;
	//sequence number
	int seq;
	//send time instant
	int send_ev;
} tcb_distributed_fd_id_t;


/**
 * Union returned by tcb_recvfrom.
 */
typedef struct tcb_distributed_action_result {
	//0-tcb_distributed_measurement_info_t
	//1-tcb_distributed_fd_id_t
	char type;
	union {
		tcb_distributed_measurement_info_t measurement_info;
		tcb_distributed_fd_id fd_info;
	};

} tcb_distributed_action_result_t;




/**
 * Opens the TCB session. This method should be called
 * first of all. 
 * In this version, only 1 session is allowed per machine.
 * 
 * @return 0 if OK, -1 if TCB is stopped, 
 *					-2 if a session is already open,
 *					-3 if other error
 */
TCB_SERVICES_LIBRARY_API int open_tcb_session(void);

/**
 * Closes the TCB session.
 * 
 * @return 0 if OK, -1 if error
 */
TCB_SERVICES_LIBRARY_API int close_tcb_session(void);

/**
 * Calls the tcb service: get_timestamp
 * 
 * Returns the local clock value (milliseconds).
 * 
 * @return Local clock value (milliseconds) if OK,
 *		   -1 if no TCB session is opened		   
 */
TCB_SERVICES_LIBRARY_API int get_tcb_timestamp(void);

/**
 * Calls the tcb service: get_global_timestamp
 * 
 * Returns the global clock value (milliseconds) at local instant specified.
 * If localInstant is 0, the present global clock value is returned.
 * 
 * @param localInstant A timestamp obtained with get_timestamp() or 0.
 * 
 * @return Global clock value (milliseconds) at local instant specified if OK
 * 		   (If localInstant is 0, the present global clock value is returned),
 *			-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int get_tcb_global_timestamp(int localInstant);

/**
 * Calls the tcb pseudo-service: get_tcb_id
 * 
 * Returns the id of the TCB we are connected to.
 * 
 * @return Id of the TCB we are connected to if OK,
 *		   -1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API char get_tcb_id(void);

/**
 * Calls the tcb service: start_local_measurement
 * 
 * Starts a local measurement at instant start_ev.
 * 
 * @param start_ev Instant that marks the begin of the measurement
 * 
 * @return Tag that identifies the measurement if OK,
 *		   -1 if no tag is available,
 *		   -2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_measurement(int start_ev);

/**
 * Calls the tcb service: end_local_measurement
 * 
 * Stops the measurement identified by tag.
 * 
 * @param tag Measurement identifier
 *
 * @return Information about measurement if OK,
 *		   tcb_measurement_info_t.end_ev=-1 if tag is invalid,
 *		   tcb_measurement_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_measurement_info_t end_tcb_local_measurement(int tag);

/**
 * Calls the tcb service: timely_exec
 * 
 * Starts the execution of an application function. 
 * TCB verifies if it's possible to timely execute it.
 * 
 * @param start_ev A time instant
 * @param delay start_ev+delay marks the exec begin
 * @param max_exec Deadline for the execution, counting since start_ev
 * @param dllName Name of the DLL file where the function is defined
 * @param funcName Function name 
 * @param wcetName Worst Case Execution Time function name
 * 
 * @return The time instant when the function ends if OK;
 *			-1 if no available entry in the executions table
 *			-2 if dll not found
 *			-3 if function not found
 *			-4 if wcet function not found
 *			-5 if it's not possible to execute function
 *			-6 if it's not possible to timely execute function
 *			-7 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API int tcb_timely_exec(int start_ev, int delay, int max_exec, char* dllName,
										 char* funcName, char* wcetName);

/**
 * Calls the tcb service: start_local_fd
 *
 * Starts a local timing failure detection.
 * 
 * @param start_ev A time instant
 * @param spec A timing failure occurs if start_ev+spec instant is reached
 *	            before a call endLocalFD
 * @param deadline Deadline for the execution of handler, if a failure occurs
 * @param dllName Name of the DLL file where the handler is defined
 * @param funcName Handler name 
 * @param wcetName Handler's Worst Case Execution Time function name
 *
 * @return Tag that identifies the timing failure detection if OK,
 *			-1 if no available entry in the localFD table,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if it's not possible to execute timing failure detection,
 *			-6 if a timing failure already occured,
 *			-7 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int start_tcb_local_fd(int start_ev, int spec, int deadline, 
											char* dllName, char* funcName, char* wcetName);

/**
 * Calls the tcb service: end_local_fd
 *
 * Ends a local timing failure detection.
 * 
 * @param tag Local timing detection identifier
 *
 * @return Information about the timing failure detection if OK,
 *			local_fd_info_t.end_ev=-1 if tag is invalid,
 *			local_fd_info_t.end_ev=-2 if no TCB session is opened
 */
TCB_SERVICES_LIBRARY_API tcb_local_fd_info_t end_tcb_local_fd(int tag);

/**
 * Waits for info from TCB distributed timing failure detection service.
 *
 * @return Information about the distributed timing failure detection if OK,
 *		   tcb_distributed_fd_info_t.from=-1 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_fd_info_t tcb_wait_info(void);

/**
 * Creates a TCB socket. 
 * At each time, there can be only one tcb socket created, so no descriptor 
 * is returned.
 *
 * @param family Address family specification
 *
 * @return 0 if OK, 
 *		   -1 if a socket error occurs.
 *		   -2 if no TCB session is opened.		   
 */
TCB_SERVICES_LIBRARY_API int tcb_socket(int family);

/**
 * Associates a local address with the TCB socket.
 *
 * @param addr Address to assign to the socket
 * @param addrlen Length of address
 *
 * @return 0 if OK, 
 *		   1 if no TCB session is opened,
 *		   SOCKET_ERROR if a socket error occurs
 */
TCB_SERVICES_LIBRARY_API int tcb_bind(struct sockaddr *addr, int addrlen);

/**
 * Calls the TCB pseudo-service: define_remote_fd_handler
 *
 * GOAL: to define the handler to be used on remote failure detection.
 *
 * PARAM:
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *		deadline Maximum exec time of failure handler (if a failure occurs)
 *
 * RETURNS: 
 *		0 if OK,
 *		-2 if dll not found,
 *		-3 if function not found,
 *		-4 if wcet function not found,
 *		-5 if no TCB session is opened.
 */
TCB_SERVICES_LIBRARY_API int define_tcb_remote_fd_handler(char* dllName, char* funcName, char* wcetName, 
							 int deadline);

/**
 * Sends data to a specific destination through the TCB socket,
 * starting a distributed measurement.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the action to be measured
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *		   SOCKET_ERROR if a socket error occurs
 */
TCB_SERVICES_LIBRARY_API int tcb_sendto(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev);


/**
 * Sends data to a specific destination through the TCB socket,
 * starting simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the sender.
 *
 * If data is not delivered to all recipients specified in destMask until send_ev+spec,
 * the specified failure handler is executed (in the sender) with the maximum
 * exec time specified.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the failure detection and measurement
 * @param dest_mask bit mask indicating packet destinations. If bit n=1,
 *				    then packet is sent to destination where TCB id=n.
 *				    E.g. dest_mask=0x7 => send to destinations with TCB ids=0,1,2.
 *					This information should be consistent with the remote IP address
 *					specified in to
 * @param spec Maximum duration starting at send_ev (for failure detection)
 * @param deadline Maximum exec time of failure handler (if a failure occurs)
 * @param dllName Name of the DLL file where the failure handler is defined
 * @param funcName Failure handler name
 * @param wcetName Failure handler Worst Case Execution Time function name
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *			SOCKET_ERROR if a socket error occurs
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API int tcb_sendtoWithFD(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev,
										char dest_mask, int spec, int deadline,
										char* dllName, char* funcName, char* wcetName);

/**
 * Sends data to a specific destination through the TCB socket,
 * starting simultaneously a distributed measurement and a distributed
 * timing failure detection.
 * If a timing failure occurs, the failure handler is executed in the recipients.
 *
 * For each recipient specified in dest_mask: if data is not delivered until send_ev+spec,
 * the failure handler specified in define_tcb_remote_fd_handler
 * is executed with the maximum exec time specified in that function.
 *
 * @param buf buffer that contains the data to be transmitted. 
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len length of buf
 * @param flags specifies the way in which the call is made
 * @param to address of the target socket
 * @param tolen length of address
 * @param send_ev the initial time instant of the failure detection and measurement
 * @param dest_mask bit mask indicating packet destinations. If bit n=1,
 *				    then packet is sent to destination where TCB id=n.
 *				    E.g. dest_mask=0x7 => send to destinations with TCB ids=0,1,2.
 *					This information should be consistent with the remote IP address
 *					specified in to
 * @param spec Maximum duration starting at send_ev (for failure detection)
 *
 * @return total number of bytes sent if OK, 
 *		   -100 if no TCB session is opened,
 *			SOCKET_ERROR if a socket error occurs,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection
 */
TCB_SERVICES_LIBRARY_API int tcb_sendtoWithRemoteFD(char* buf, int len, int flags, 
										struct sockaddr *to, int tolen, int send_ev,
										char dest_mask, int spec);

/**
 * Receives data from the TCB socket. This method ends a distributed
 * measurement but doesn't end a distributed failure detection. For that, 
 * call end_tcb_distributed_action() with the tcb_distributed_fd_id_t returned by 
 * this method.
 * 
 * @param buf buffer for the incoming data.
 *			  ATTENTION: The last TCB_DATA_SIZE bytes of the buffer 
 *						 will be used by TCB
 * @param len (IN:length of buf)(OUT:bytes received)
 * @param flags specifies the way in which the call is made
 * @param from pointer to a buffer that will hold the source address upon return
 * @param fromlen (IN:pointer to the size of the from buffer)
 *				  (OUT:actual size of from buffer)
 * 
 * @return The return value is a union. Its value depends on the service 
 *		   called by the sender. 
 *		   If it's only a distributed measurement, a 
 *		   tcb_distributed_measurement_info_t is returned.
 * 		   If it's a distributed failure detection, a 
 *		   tcb_distributed_fd_id_t is returned.
 *
 *		   If there is no TCB session opened, 
 *		   tcb_distributed_action_result_t.type=-100.
 *         
 *		   If a socket error, tcb_distributed_action_result_t.type=SOCKET_ERROR.
 * 
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_action_result_t tcb_recvfrom(
										  char* buf, int* len, int flags, 
										  struct sockaddr* from, int* fromlen);

/**
 * Ends a distributed failure detection and the associated measurement.
 * 
 * @param dfi Id of the action to be ended.
 * 
 * @return Info about the result of the distributed measurement if OK,
 *		   tcb_distributed_measurement_info_t.end_ev=-100 if there is no 
 *		   TCB session opened,
 *		   -1 if a socket error occurs.
 */
TCB_SERVICES_LIBRARY_API tcb_distributed_measurement_info_t 
			end_tcb_distributed_action(tcb_distributed_fd_id_t dfi);