/**
 * TCB for Windows CE 3.0
 *
 * This header file defines the structures used by the TCB
 *
 * @author Paulo Sousa
 * (based on RTAI TCB implementation by Pedro Martins)
 * @date 18/Nov/2002
 */

#ifndef STRUCTURES_TCB_H
#define STRUCTURES_TCB_H

/*Msg types*/
#define DATA 1

//stores a machine name and it's IP address
typedef struct machine {
	char name[30];
	char ip[16];
	int port;
} machine_t;
//stores machines names and addresses
machine_t machines[NUMBER_OF_MACHINES];


//stores information that goes in each position of data_msg_hdr.RoundTripInfo
typedef struct {
	//=snd_time - last_rcv_info[<destination>].timestamp
 	int rndval;  	
 	//=last_rcv_info[<destination>].snd_time
	int snd_time;
} roundtripVal;

//stores information used by the distributed timing failure detection service.
//This information is packed in a data msg.
typedef struct {
	//0->it's a request, 1->it's a response pt2pt (only for the service requester)
	//2->it's a reponse for all machines.
	char type;
	union {
		//in a request indicates if it's a remote failure detection
		bool remote_fd;
		//in a response indicates if a timing failure has occured.
		bool fail;
	};
	//in a request it has no meaning.
	//in a reponse indicates if the lateness degree is available
	bool lateness_avail;
	//in a request it stores the bit mask that specifies recipients.
	//in a response it stores the id of the TCB that requested the service. 
	char destination;
	union {
		//in a request it stores the time remaining for the failure.
		int time_to_failure;
		//in a reponse it stores the lateness degree, if available
		int lateness_degree;
	};
	//in a request/response it stores the sequence number that identifies
	//the timing failure detection operation
	int seq;
} info_field_t;


//header of a data msg (defined below)
typedef struct data_msg_hdr {
		//msg type (=DATA)
		char type;
		//number of info fields
		char num_info_fields;
		//source machine id
		unsigned char machine_src_id;
		//sequence number
		int seq;
		//send time instant
		int snd_time;
		//global clock value
		int vclock;
		//Information for the mechanism that measures msg delivery times
		roundtripVal RoundTripInfo[NUMBER_OF_MACHINES];	
		
} data_msg_hdr_t;

//Maximum number of info fields for the distributed FD service
#define MAX_NUM_INFO_FIELDS ((MAX_PACKET_SIZE-sizeof(data_msg_hdr_t))/sizeof(info_field_t))

//Maximum number of simultaneous distributed FD services
#define MAX_DISTRIBUTED_FD_SERVICES (MAX_NUM_INFO_FIELDS/(NUMBER_OF_MACHINES+1))

//Maximum size of distributed FD send info table
#define MAX_SEND_INFO_TABLE_SIZE MAX_DISTRIBUTED_FD_SERVICES

//Maximum size of distributed FD receive info table
#define MAX_RCV_INFO_TABLE_SIZE (MAX_DISTRIBUTED_FD_SERVICES*NUMBER_OF_MACHINES)

//Maximum size of distributed FD wait info table. 
//The multiplication by a factor of 3 is a safety margin.
#define MAX_WAIT_INFO_TABLE_SIZE \
				(MAX_DISTRIBUTED_FD_SERVICES*NUMBER_OF_MACHINES*3)

//Maximum size of distributed FD receive payload table. It stores the msg
//received on the payload channel between 2 receive cycles. 
//If during TCB operation this table fills up, TCB simulates
//omissions on the payload channel to maintain system consistency.
#define MAX_RCV_PLOAD_TABLE_SIZE 100 


//stores a data msg used by the network access protocol.
typedef struct data_msg {
	//This header preceeds the info_fields  
	data_msg_hdr_t hdr;
	//msg content: info_fields
	char data[MAX_NUM_INFO_FIELDS*sizeof(info_field_t)];
} data_msg_t;

//stores a token msg used by the network access protocol
typedef struct token_msg {
	//msg type (=TOKEN)
	char type;
	//source machine id
	unsigned char machine_src_id;
	//sequence number
	int seq;
} token_msg_t;


//stores information about the last data message received from each node
typedef struct {
	//sequence number
	int seq;
	//reception time instant
	int timestamp;
	//send time instant (measured in the sender clock)
	int snd_time;
} msg_info_type;


//stores the information stored in send queue and receive queue
typedef struct data_msg_with_timestamp {
	//data msg to send/received
	data_msg_t dm;
	//timestamp used to build data info/time of receipt
	int timestamp;
} data_msg_with_timestamp_t;



//stopMeasurement return value
typedef struct {
	int end_ev;
	int procdelay;
} measurement_info_t;


//used in fail_info_to_app and in wait_info. Stores info about a failure.
typedef struct fail_info {
	//0-no failure, 1-failure, 2-no info
	char fail;
	//lateness is not available (FALSE), or available(TRUE)
	bool lateness_avail;
	//lateness degree if available
	int lateness_degree;
} fail_info_t;

//for memory mapping between TCB DistFD service and 
//wait_info() function implemented in TCB DLL
typedef struct fail_info_to_app {
	//if msg was sent on broadcast or not
	char broadcast;
	//msg sender id (tcb id)
	char from;
	//msg sequence number (FD context)
	int seq;
	//info about failure in each machine
	fail_info failure_info[NUMBER_OF_MACHINES];
} fail_info_to_app_t;		


//for memory mapping between TCB services DLL and TCB
//Note: This info is shared with the DLL
#define GET_TIMESTAMP 0
#define START_MEASUREMENT 1
#define END_MEASUREMENT 2
#define TIMELY_EXEC 3
#define START_LOCAL_FD 4
#define END_LOCAL_FD 5
#define GET_TCB_ID 6
#define START_DISTRIBUTED_FD 7
#define MSG_RCV_ON_PAYLOAD 8
#define GET_GLOBAL_TIMESTAMP 9
#define DEFINE_REMOTE_HANDLER 10
#define START_DISTRIBUTED_REMOTE_FD 11

typedef struct service_request_response {
	//0-empty, 1-service request, 2- service response
	char type;
	union {
		//request
		struct {
			//defines above
			char service_id;
			//service input parameters
			union {
				//getGlobalTimestamp input value
				int localInstant;
				//startMeasurement input value
				int start_ev;
				//endMeasurement and endLocalFD input value
				int tag;
				//timelyExec input value
				struct {
					//a time instant
					int start_ev; 
					//start_ev+delay marks the exec begin
					int delay; 
					//deadline for the execution, counting since start_ev
					int max_exec; 
					//DLL file where the function is defined
					wchar_t dllName[50]; 
					//function name
					wchar_t funcName[50]; 
					//Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} execArguments;
				//startLocalFD input value
				struct {
					//start instant of the observed execution
					int start_ev;
					//specified execution duration
					int spec;
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} localFDArguments;
				//startDistributedFD and (partially)startDistributedRemoteFD input value
				struct {
					//bit mask indicating packet destinations
					wchar_t destMask;
					//start instant of the observed execution
					int send_ev;
					//specified execution duration
					int spec;
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
				} distFDArguments;
				//msgRcvOnPayload input value
				struct {
					int seq;
					wchar_t from;
					int time_of_receipt;
				} msgRcvOnPayloadArguments;
				//defineRemoteHandler input value
				struct {
					//DLL file where the handler is defined
					wchar_t dllName[50]; 
					//handler name
					wchar_t funcName[50]; 
					//handler's Worst Case Execution Time function name
 					wchar_t wcetName[50]; 
					//maximum time allowed for the execution of the handler
					//(if a failure occurs)
					int deadline;
				} remote_handler;
			};
		} request;

		//response
		struct {
			//defines above
			char service_id;
			//service return values
			union {
				//getTimestamp and getGlobalTimestamp return value	
				int value;
				//startMeasurement and startLocalFD return value	
				int tag;
				//endMeasurement return value
				measurement_info_t mi;
				//timelyExec return value
				int end_ev;
				//endLocalFD return value
				struct {
					//when the detection was turned off
					int end_ev;
					//end_ev - start_ev
					int procdelay;
					//a timing failure ocurred?
					bool faulty;
				} endLocalFDRet;
				//getTCBId return value
				int id;
				//startDistributedFD return value
				int seq;
				//msgRcvOnPayload return value
				bool deliver;
				//defineRemoteFDHandler return value 
				int ret;
			};
		} response;
	};
} service_request_response_t;


/*types of the functions executed by TCB services*/
//handler for local services
typedef void (*functype)(void);
//handler for distributed services -> receives <from.seq>
typedef void (*distfunctype)(char,int);
//handler WCET 
typedef int (*wcettype)(void);




//timely execution service table entry
//used for communication between the function that receives
//the service and the task that executes the app function
typedef struct {
	//function to be executed
	functype func;
	//DLL handle
	HMODULE hdll;
	//func Worst Case Execution Time
	int wcet;
	//func start delay
    int delay;
	//func start instant
    int start_ev;
	//func desired maximum execution time 
    int max_exec;
} exec_info_t;


//local timing failure detection service table entry
//used for communication between the function that receives
//the service and the task that executes the handler if a local failure
//occurs
typedef struct {
	//function to be executed if a failure occurs
	functype func;
	//DLL handle where the function resides
	HMODULE hdll;
	//thread handle that will monitor if failure occurs
	HANDLE task_id;
	//indicates if a failure occured
	bool faulty;
	//start instant of the observed execution
    int start_ev;
	//specified execution duration
    int spec;
	//maximum time allowed for the execution of func
	//(if a failure occurs)
    int deadline;
} local_fd_info_t;


//used for communication between the function that receives
//the service and the task that executes the handler if a 
//distributed failure occurs
typedef struct {
	//delay before starting failTask (currently not used)
	int release_time;
	//maximum time allowed for the execution of funcptr
	//(if a failure occurs)
	int deadline;
	//function to be executed if a failure occurs
	distfunctype funcptr;
	//failure instant (absolute value!)
	int failure_at_time;
	//msg source
	char from;
	//msg seq number
	int seq;
} dist_fd_info_t;


//send info table entry:
//stores a distributed FD request that was not yet disseminated
//in the control channel (pending sends)
typedef struct {
	//initial instant of the observed execution
	int time_of_generation;
	//maximum duration starting at time_of_generation
	int spec;
	//sequence number assigned by TCB
	int seq;
	//destination mask: identifies recipients
	char destMask;
	//indicates if it's a remote FD
	bool remote_fd;
} send_info_t;


//wait info table entry:
//stores a distributed FD operation until it being delivered to
//application.
typedef struct {
	//failure instant (absolute value!)
	int failure_at_time;
	//sequence number assigned by TCB
	int seq;
	//number of msg recipients
	char n_dests;
	//msg sender id
	char from;
	//destination mask: identifies recipients
	char destMask;
	//indicates if was received a failure notification
	bool any_fail;
	//information about the FD operation in each of the recipients
	fail_info_t failure_info[NUMBER_OF_MACHINES];
	//thread handle that will monitor if failure occurs
	HANDLE task_id;
	//indicates if it's a remote FD
	bool remote_fd;
} wait_info_t;


//receive info table entry:
//stores a distributed FD request received
//in the control channel until the result being known
typedef struct {
	//failure instant (absolute value!)
	int failure_at_time;
	//request sender id
	char from;
	//sequence number assigned by TCB
	int seq;
	//destination mask: identifies recipients
	char destMask;
	//0-no failure, 1-failure, 2-no info
	char fail;
	//lateness is not available (FALSE), or available(TRUE)
	bool lateness_avail;
	//lateness degree if available
	int lateness_degree;
	//indicates if it's a remote FD
	bool remote_fd;
} rcv_info_t;


//receive payload info table entry:
//stores information about msg received in the payload channel and
//whose failure info was not yet received by TCB in the control channel
typedef struct {
	//reception instant
	int time_of_receipt;
	//msg sender id
	char from;
	//sequence number assigned by sender TCB 
	int seq;
	//time instant that indicates when an entry can be "cleaned"
	//by the garbage collector
	int clean_up_after;
} rcv_payload_info_t;



//stores information about a global clock
typedef struct {
	//value
	int vclock;
	//error
	double error;
	//instant of last synchornization
	int time_of_sync;
} global_clock_info_t;


#endif